

import onearmap


from numpy import ones, zeros, array, mean, std
from fcmp import fcmp
from numpy.random import rand, randint

def Ber(p):
	"""Simuates from a Bernoulli distribution with prob. p."""
	if (rand(1)[0] < p):
		return 1
	else:
		return 0




class PyADeBay:
	"""Class to implement the approximate calculations of paper
	   Christen, Muller, Whaten and Wolf (2004), ``A Bayesian Randomized Clinical Trial:
	   A Decision Theoretic Sequential Design'', Canadian Journal of Statistics, 32(4), 387-402,
	   using binary response ONLY R=2 and the one-arm approximation of
	   Akongo Orawuo, L. and Christen, J.A. (2008), "Bayesian Sequential Analysis for
	   Multiple-arm Clinical Trials", Statistics and Computing, (to appear)."""


	def __init__( self, k, N, LE, pp, al, be, alpha=0):
		self.N = N
		self.k = k
		if (alpha == 0):
			self.alpha = 1.0/(self.N + 1)
		else:
			self.alpha = alpha
		
		self.LE = LE.copy()
		self.pp = pp.copy()

		self.al = al.copy()
		self.be = be.copy()
		
		self.tr = onearmap.OneArmAp( k=k, N=N, a=self.GetUt( 0, [0, 0]), b=self.GetUt( 1, [0, 0]), al=al, be=be)

		self.uts = [[0,0],[1,0],[0,1],[1,1]]
		self.ContU = zeros(( 16, self.k))
		self.StopU = zeros(( 16, self.k))

	
	def GetUt( self, r, s):
		"""Gives the utilities of response r= 0, 1,
		   combinations of s= [s1, s2].  s_i= 0,1; low, high:
		   return LE[ s[0], r] * pp[ s[1], :]
		   This creates the 16 possible utility functions:
		   a=GetUt( 0, [ 0 or 1, 0 or 1]), b=GetUt( 1, [ 0 or 1, 0 or 1])"""
		
		return self.LE[ s[0], r] * self.pp[ s[1], :]


	def EvalAllUt( self, sn, ss, printprogress=False):
		"""Evaluate expected utilities for all 16 utility values,
		   the results are save in self.ContU and self.StopU.  The history is:
		   sn number of each treatment uses, and ss successes for each treatment."""
		
		utnum = 0
		for s1 in self.uts:
			for s2 in self.uts:
				if (printprogress):
					print "\nEvaluating one arm approximatioon for utlities %d of 15" % utnum
				self.tr.ChangeUt( a=self.GetUt( 0, s1), b=self.GetUt( 1, s2))
				
				tmp = self.tr.Evaluate( sn=sn, ss=ss, printprogress=printprogress)
				self.StopU[ utnum, :] = tmp[3]
				self.ContU[ utnum, :] = tmp[4]
				utnum += 1
	
	
	def Dominates( self, d, j, i, printprogress=False):
		"""Does treatment i dominates treatment j?
		   For this if d=0 we use self.ContU otherwise
		   we use self.StopU; return False or True."""
		
		if (printprogress):
			print "\nDoes treatment %d dominates treatment %d?" % (i,j) 
		rt = True
		utnum = 0
		while (rt & (utnum < self.k)):
			if (d == 0):
				rt = (fcmp(self.ContU[utnum,j], self.ContU[utnum,i]) == -1) ## x1 < x2
				if (printprogress):
					if not(rt):
						print "No: utnum= %d, %f !< %f" % ( utnum, self.ContU[utnum,j], self.ContU[utnum,i])  
			else:
				rt = (fcmp(self.StopU[utnum,j], self.StopU[utnum,i]) == -1) ## x1 < x2
				if (printprogress):
					if not(rt):
						print "No: utnumut= %d, %f !< %f" % ( utnum, self.ContU[utnum,j], self.ContU[utnum,i])  
			utnum += 1
		if (printprogress & rt):
			print "Yes."
		
		return rt


	def Decision( self, sn, ss, printprogress=False):
		"""Provides de decision, stop or continue,
		   and the non-dominated set for randomization.  The history is:
		   sn number of each treatment uses, and ss successes for each treatment."""
		
		self.EvalAllUt( sn, ss, printprogress=printprogress)
		
		if (printprogress):
			print "\nEvaluating non-domminating sets."
		### Evaluate whether to stop or continue.
		d = 1
		utnum = 0
		while ((d == 1) & (utnum < self.k)):
			if (fcmp(max(self.StopU[utnum,:]), max(self.ContU[utnum,:])) == -1): # x1 < x2
				if (printprogress):
					print "Utility %d prescribes continuation." % utnum
				d = 0
			utnum += 1
		if (printprogress & (d == 1)):
			print "Stopping is recommended." 
		

		### Find the non-dominated set for randomization
		### Algorithm taken from Rios-Insua et al. (1997), J. Stat. Comp. Sim., 57(1), 197-218.
		Dom = zeros(self.k)
		for i in range(self.k-1):
			if (Dom[i] == 0):
				for j in range( i+1, self.k, 1):
					utnum = 0
					if (self.Dominates( d, j, i, printprogress=printprogress)):
						Dom[j] = 1 # j treatment is dominated by i treatment
					else:
						if (Dom[i] == 0):
							if (self.Dominates( d, i, j, printprogress=printprogress)):
								Dom[i] = 1 # i treatment is dominated by j treatment
		
		nondom = []
		for t in range(self.k):
			if (Dom[t] == 0):
				nondom = nondom + [t]

		return [ d, nondom]


	def TrueNonDomSet( self, trueprobs):
		"""Calculate true non-dominated set, only v utility function is used."""

		###  Calculate utilities and save them in self.StopU
		utnum = 0
		for s1 in self.uts:
			for s2 in self.uts:
				self.tr.ChangeUt( a=self.GetUt( 0, s1), b=self.GetUt( 1, s2))
				for t in range(self.k):
					self.StopU[ utnum, t] = self.tr.UStopTrueProb( t, trueprobs[t])
				utnum += 1

		### Find the non-dominated set for randomization
		### Using the utilities in self.StopU
		d = 1
		Dom = zeros(self.k)
		for i in range(self.k-1):
			if (Dom[i] == 0):
				for j in range( i+1, self.k, 1):
					utnum = 0
					if (self.Dominates( d, j, i)):
						Dom[j] = 1 # j treatment is dominated by i treatment
					else:
						if (Dom[i] == 0):
							if (self.Dominates( d, i, j)):
								Dom[i] = 1 # i treatment is dominated by j treatment

		nondom = []
		for t in range(self.k):
			if (Dom[t] == 0):
				nondom = nondom + [t]

		return nondom



	def SimulateTr( self, trueprobs, sn=0, ss=0, printprogress=False):
		"""Simultaes a trial using the true probabilities trueprobs.  The history is:
		   sn number of each treatment uses, and ss successes for each treatment.
		   If sn=0 (ss=0) these will be set to zeros, ie. start with no history."""
		
		if (ss == 0):
			ss = [0]*self.k
			sn = [0]*self.k
		
		### Check if we can start the trial 
		n = 0
		res = self.Decision( sn=sn, ss=ss) 
		
		d = res[0]
		nds = [res[1]]  # To save non-dom. sets
		tt = []       # To save the sequence of treatments
		xx = []       # To save the sequence of responses
		while (d == 0):
			n += 1
			### Randomize treatment within the non-dominated set.
			i = randint(0, len(res[1]))
			t = res[1][i]
			
			### Simulate response and add result to sn and ss
			r = Ber(trueprobs[t])
			sn[t] += 1
			ss[t] += r
			tt = tt + [t]
			xx = xx + [r]

			if (printprogress):
				print "Patient %d, with treatment %d and response %d" % ( n, t, r)
			res = self.Decision( sn=sn, ss=ss)
			
			d = res[0]
			nds = nds + [res[1]]
			if (printprogress):
				print "      Non-dominated set: ", res[1]
		
		return [ n, nds, tt, xx, sn, ss]



	def OpCa( self, trueprobs, T, printprogress=False):
		"""Runs a set of T trial simulations with the true probabilities trueprobs."""

		TrueNDS = set(self.TrueNonDomSet(trueprobs)) 
		
		res = []
		nn  = []
		### We will set NDS[i,t]=1, if treatment t was included in the final non-dom. set
		NDS = zeros((T,3))
		sns = zeros((T,3))
		
		for i in range(T):
			if (printprogress):
				print "Running simulation %4d of %4d" % ( i+1, T)
			out = self.SimulateTr(trueprobs=trueprobs)

			FinalNDS = set(out[1][out[0]])
			for t in range(self.k):
				if (t in FinalNDS):
					NDS[i,t]=1
			### Is the Final non-dom. set contained in the true non-dom. set?
			if (FinalNDS <= TrueNDS):
				res += [1]				
				if (printprogress):
					print "True non-dom. ", TrueNDS, " contains final non-dom. ", FinalNDS 
			else:
				res += [0]				
				if (printprogress):
					print "True non-dom. ", TrueNDS, " does not contain final non-dom. ", FinalNDS 
			nn  += [sum(out[4])]
			sns[i,:] = out[4]
		
		self.OPCA = dict( T=T, N=self.N, trueprobs=trueprobs, res=res, nn=nn, sns=sns, NDS=NDS)

		if (printprogress):
			self.PrintOPCA()
			return 1
		else:
			return self.OPCA


	def PrintOPCA( self, OP=0):
		"""Print an outline of the Operating Characteristics OP.
		   Typically if OP == 0, OP = self.OPCA, the Op Cha. saved by the method OpCa."""
		
		if (OP == 0):
			OP = self.OPCA
		
		print "\n\n****** Operating Characteristics ******"
		print "Number of iterations: %d, trial length: %d\n" % (OP['T'],OP['N'])
		print "Trt     True Prob.  Mean n (std. dev)   %RS"
		print "--------------------------------------------"
		for t in range(self.k):
			percp = 100 * OP['trueprobs'][t]
			mnn   = OP['sns'][:,t].mean()
			stdn  = OP['sns'][:,t].std()
			RS    = 100 * sum(OP['NDS'][:,t])/OP['T']
			print "%3d       %5.1f    %4.1f (%4.1f)        %5.1f"  % ( t, percp, mnn, stdn, RS)
		print "--------------------------------------------"
		print "Mean number of patientes enrrolled: %5.1f (%4.1f)" % ( mean(OP['nn']), std(OP['nn']))

		stopearly = 0
		for i in range(OP['T']):
			if (OP['nn'][i] < OP['N']):
				stopearly += 1
		se     = (100.0 * stopearly) / OP['T'] 
		success = (100.0 * sum(OP['res'])) / OP['T']
		print "Perc. of trials stopped early: %5.1f" % se
		print "Perc. of times the final non-dom. set\n   is included in the true non-dom. set: %5.1f" % success
		
		return 1





