#
#  onearmap.py
#  
#


from numpy import ones, zeros, array

class OneArmAp:
	"""Class to implemente the one-arm approximation of
	   Orawou  and Christen. 
	   k = Number of treatments.
	   N = Horizon.
	   a = array with utilities of failure for each treatment.
	   b = array with utilities of success for each treatment.
	    NB: THESE UTILITIES ARE ASSUMMED NON-NEGATIVE, typically these are QALY's.
	   al, be = Beta( al[t], be[t]) prior for prob. of success for treatment t.
	   alpha = alpha factor, if ==0, then al=1/(N+1).
	   
	   The member method 'Evaluate' provides the required output."""


	def __init__( self, k, N, a, b, al, be, alpha=0):
		self.N = N
		self.k = k
		if (alpha == 0):
			self.alpha = 1.0/(self.N + 1)
		else:
			self.alpha = alpha
		
		self.a = a.copy()
		self.b = b.copy() 
		self.al = al.copy()
		self.be = be.copy()
	
	
	def ChangeUt( self, a, b):
		"""a = array with utilities of failure for each treatment.
		   b = array with utilities of success for each treatment.
	       These Utilities are assumed non-negative, typically in [0,1]."""
		self.a = a.copy()
		self.b = b.copy() 


	def v( self, t, x):
		"""The utility of using treatment t and obtaining response x
		   t = 0,1,...,k-1, and x = 0,1 ."""
		if (x == 0):
			return self.a[t]
		else:
			return self.b[t]


	def UStopHistory( self, n, tt, xx, t1, p1):
		"""Utility of stopping:
		   stop at n, with tt the history of treatments and with xx responses,
		   next treatment = t1, pred. prob. of success
		   p1 (and of failure p0 = 1-p1)."""
		S = 0.0
		alpha = self.alpha
		N = self.N
		for i in range(n):
			S += self.v( tt[i], xx[i])
		
		return ((1.0-alpha)/N)*S + (alpha + (1.0-alpha)*(N-n)/N)*(self.v( t1, 0)*(1.0-p1) + self.v( t1, 1)*p1)


	def Pred( self, t, nt, st):
		"""Predictive probability of success for treatment t,
		   nt times used with st sucesses."""
		return (self.al[t] + st)/(nt + self.al[t] + self.be[t])


	def UStop( self, sn, ss, t1):
		"""Utility of stopping:
		   stop at n, with sn number of each treatment uses, and ss successes for each treatment,
		   next treatment = t1, pred. prob. of success = Pred( t1, sn[t1], ss[t1])."""
		S = 0.0
		alpha = self.alpha
		N = self.N
		n = sum(sn)
		for t in range(self.k):
			S += self.v( t, 1)*ss[t] + self.v( t, 0)*(sn[t]-ss[t])
		p1 = self.Pred( t1, sn[t1], ss[t1])
		
		return ((1.0-alpha)/N)*S + (alpha + (1.0-alpha)*(N-n)/N)*(self.v( t1, 0)*(1.0-p1) + self.v( t1, 1)*p1)


	def UStopTrueProb( self, t1, p1):
		"""Utility of stopping:
		   stop at n, with sn number of each treatment uses, and ss successes for each treatment,
		   next treatment = t1, pred. prob. of success = Pred( t1, sn[t1], ss[t1])."""
		return (self.v( t1, 0)*(1.0-p1) + self.v( t1, 1)*p1)


	def USt( self, sn, ss):
		"""Utility of stopping, max_t1 UStop( sn, ss, t1)."""
		maxU = 0.0
		for t in range(self.k):
			maxU = max( maxU, self.UStop( sn, ss, t))
		
		return maxU


	def OneArmUt( self, sn, ss, t):
		"""Utility of the one arm approximation;
			assuming only treatment t will be used for continuation
			or the best treatment for stopping."""
		N = self.N
		n = sum(sn)
		re = N-n+1
		#print "N= %d, n= %d, re= %d\n" % ( N, n, re)
		### Matrix to hold all caluclations.  Only the diagobal is used, a better,
		### sparse matrix mathod, could be used.
		Ut = zeros(( re, re))
		vecs = zeros(self.k)
		vecn = zeros(self.k)
		
		### We fill the last column with utilities of stopping
		j = re-1
		vecn[t] = j
		for s in range(j+1):
			#print "j= %d, s= %d\n" % (j,s)
			vecs[t] = s  ### Additional responses
			Ut[ s, j] = self.USt( sn+vecn, ss+vecs)

		### Backward induction, we go backwards for each column
		for j in range( re-2, -1, -1):
			vecn[t] = j
			### Then upwards for each additional response
			for s in range(j+1):
				#print "j= %d, s= %d\n" % (j,s)
				vecs[t] = s  ### Additional responses
				p1 = self.Pred( t, sn[t]+j, ss[t]+s)
				Ucont = p1*Ut[ s+1, j+1] + (1.0-p1)*Ut[ s, j+1]
				Ustop = self.USt( sn+vecn, ss+vecs)
				Ut[ s, j] = max( Ucont, Ustop)
		
		if (re <= 1):
			return array([ Ut[0,0], Ut[0,0], Ut[0,0] ])
		else:
			return array([ Ut[0,0], Ustop, Ucont])


	def Evaluate( self, sn, ss, printprogress=False):
		"""Calculate the utility of continuing using the one arm approximation,
		   we calculate OneArmUt( sn, ss, t) for each t, the history is:
		   sn number of each treatment uses, and ss successes for each treatment.
		   Also the Utility of stopping with each t is provided.  The return value is a list with:
		   [ decison, treatment to follow, over all utility, ut. of stopping, and ut. of cont.]"""
		
		maxContU = 0.0
		tCont = 0
		maxStopU = 0.0
		tStop = 0
		ContU = zeros(self.k)
		StopU = zeros(self.k)
		for t in range(self.k):
			if (printprogress):
				print "Analyzing treatment t= %d" % t
			ContU[t] = self.OneArmUt( sn, ss, t)[2]
			StopU[t] = self.UStop( sn, ss, t)
			if (maxContU <= ContU[t]):
				maxContU = ContU[t]
				tCont = t
			if (maxStopU <= StopU[t]):
				maxStopU = StopU[t]
				tStop = t
		
		if (maxStopU <= maxContU):  ### If equal continue!!
			d = 0
			t = tCont ### with treatment t
			Ut = maxContU ### and with this utility.
		else:      ### Else stop
			d = 1
			t = tStop ### with treatment t
			Ut = maxStopU ### and with this utility.
		
		return [ d, t, Ut, StopU, ContU]
				





